import logging
import unicodedata
import translitcodec

from flatland import schema
from json import dumps

from . memo import Memo
from . tools import geoprint

from .exc import (
    InvalidTermError,
    )

logger = logging.getLogger(__name__)

def _schema_name(schema):
    return "%s.%s" % (schema.__module__, schema.__class__.__name__)


def _prefix(prefix, value=None):
    if prefix is None:
        return
    return ("%s:%s" % (prefix, value)) if value else prefix


def _normalize(value):
    return unicodedata.normalize(
        'NFKC', unicode(value).strip().lower()).encode('utf-8')


_use_schema = object() # marker says use schema term generator


class Schema(schema.Form):

    language = None
    """The overarching language for this schema.

    If None, no language features are applied to Text elements unless
    they explicitly provide their own language.

    If a two letter language code, is used as the language for Text
    elements in this schema unless they specifically provide their
    own.
    """

    facet_prefix = 'facet'
    """The default prefix that will be used for storing and searching
    for facets.
    """

    ignore_invalid_terms = False
    """When True, invalid terms will be skipped with a warning.
    When False (default) an invalid term in a document raises InvalidTermError.
    """

    def update_by_object(self, obj):
        """Updates fields with an object's attributes.

        :param obj: any object

        Sets fields on *self*, using as many attributes as possible from
        *obj*.  Object attributes that do not correspond to field names are
        ignored.
        """
        for name, element in self.iteritems():
            if element.getter is not None:
                self[name] = element.getter(self, obj, element)
            elif hasattr(obj, name):
                self[name] = getattr(obj, name)

    @property
    def memo(self):
        """
        Walk the children, indexing each one with a handler into a
        memo object.
        """
        lang = None
        try:
            root_lang = self.root.get('language')
            lang = root_lang.value or root_lang.default
        except KeyError:
            lang = self.language

        self._memo = Memo()
        self._memo.set_lang(lang)
        self._handle_children(self)
        self._memo.data = dumps((_schema_name(self), self.flatten()))
        return self._memo

    def _handle_children(self, parent):
        for el in parent.all_children:
            if el.index:
                if isinstance(el, String):
                    handler = self._handle_string
                elif isinstance(el, Integer):
                    handler = self._handle_integer
                elif isinstance(el, Float):
                    handler = self._handle_float
                elif isinstance(el, Boolean):
                    handler = self._handle_boolean
                elif isinstance(el, Date):
                    handler = self._handle_date
                elif isinstance(el, DateTime):
                    handler = self._handle_datetime
                elif isinstance(el, Text):
                    handler = self._handle_text
                elif isinstance(el, NumericRange):
                    handler = self._handle_numericrange
                elif isinstance(el, Location):
                    handler = self._handle_location
                elif isinstance(el, List):
                    handler = self._handle_children
                elif isinstance(el, Dict):
                    handler = self._handle_children
                elif isinstance(el, Array):
                    handler = self._handle_children
                else:
                    raise TypeError("Unknown element %s" % el)

                value = None
                try:
                    value = handler(el)
                except InvalidTermError, e:
                    if self.ignore_invalid_terms:
                        logger.warning('Invalid term ignored: %r' % el)
                    else:
                        raise

                if value is not None:
                    if el.facet:
                        if el.name:
                            term = el.name.lower()
                            self._memo.add_term(
                                _prefix(self.facet_prefix, term),
                                True)
        return True

    def _handle_scalar(self, term, value, element, type=None):
        name = element.flattened_name()
        memo = self._memo
        if term:
            if element.lower:
                term = term.lower()
            if element.prefix:
                memo.add_term(_prefix(name, term), element.boolean, element.wdf_inc)
            else:
                memo.add_term(term, element.boolean, element.wdf_inc)
            if element.sortable:
                memo.add_value(name, value, type)
            return value

    def _handle_string(self, element):
        term = element.u
        value = element.value
        if value:
            return self._handle_scalar(term, value, element, 'string')

    def _handle_integer(self, element):
        term = element.u
        value = element.value
        if value:
            return self._handle_scalar(term, value, element, 'integer')

    def _handle_float(self, element):
        # TODO:mp floats are currently storage only
        pass

    def _handle_boolean(self, element):
        value = 'true' if element.value else 'false'
        if value:
            return self._handle_scalar(value, value, element, 'integer')

    def _handle_date(self, element):
        if element.value:
            term = element.value.strftime(element.term_format)
            value = element.value.strftime(element.value_format)
            return self._handle_scalar(term, value, element, 'date')

    def _handle_datetime(self, element):
        if element.value:
            term = element.value.strftime(element.term_format)
            value = element.value.strftime(element.value_format)
            return self._handle_scalar(term, value, element, 'datetime')

    def _handle_text(self, element):
        value = element.value
        if value is None:
            return
        name = element.flattened_name()
        memo = self._memo

        if element.language is _use_schema:
            lang = memo.lang
        else:
            lang = element.language

        if element.translit:
            value = value.encode(element.translit)

        if element.string:
            if value:
                if element.lower:
                    term = value.lower()
                if element.string_prefix:
                    memo.add_term(_prefix(element.string_prefix, term),
                                   element.boolean, element.wdf_inc)
                else:
                    memo.add_term(term, False, element.wdf_inc)

        if element.sortable:
            memo.add_value(name, value, 'string')

        value = _normalize(value)
        prefix = None
        if element.prefix:
            prefix = _prefix(name)
        memo.add_text(value, prefix, lang,
                      element.positions,
                      element.stem,
                      element.stop,
                      element.spelling,
                      element.wdf_inc,
                      element.position_start)
        return value

    def _handle_numericrange(self, element):
        maxv = element['high'].value or 0
        minv = element['low'].value or 0
        if minv == maxv == 0:
            return

        step = element.step

        if minv < step:
            minv = 0
        else:
            minv = ((minv / step) * step)

        maxv = (((maxv + step) / step) * step)

        for i in xrange(minv, maxv, step):
            val = "%s_%s" % (i, i + step)
            val = val.lower()
            self._memo.add_term(_prefix(element.name, val), True, 
                                element.wdf_inc)
        return True

    def _handle_location(self, element):
        memo = self._memo
        h = 'loc_' + element.hash(element.radians)
        memo.add_term(h, element.boolean, element.wdf_inc)
        if element.sortable:
            memo.add_value(element.name, h, 'location')
        return True


class _BaseElement(object):

    index = True
    """If True, generate terms and values for this element.

    If False, store the flattened element in the document record, but
    do not generate terms or values.
    """

    getter = None
    """Getter function to calculate value from a given object.

    Use by Schema.update_by_object(obj) to overide the default getattr
    behavior and allow a schema author to provide custom logic.  If
    None, a getattr policy is used.

    Getter must have a two argument signature for (schema, obj).  The
    property decorator can be used to decorate a method that becomes
    this attribute.
    """

    prefix = True
    """If True, generate a prefix from the flattened name of the element.

    If False, the element is indexed with no prefix.
    """

    boolean = True
    """If True, then tell Xapian to treat this field as a boolean, not
    a probablistic term.

    Most elements default as boolean, text elements do not.
    """

    facet = False
    """Whether to generate facet terms for this elemement, or not.

    If True, the term 'facet:name' will be generated with the elements
    flattened name and added along with the elements other terms.

    If False, no faceting term is generated.
    """

    sortable = False
    """Whether to sort on the element or not.

    If True, a value slot in the document will be used to store the
    sortable value.
    """

    lower = True
    """Whether to lowercase the value when indexing the element's term.
    """

    wdf_inc = 1
    """How much to increment the within document frequency for terms
    generated by this element.
    """

    @classmethod
    def property(cls, fn):
        """Decorator for wrapping a schema class method getter with an
        element.

        Wrapped method must have the same signature expected by
        'getter'.
        """
        cls.getter = staticmethod(fn)
        return cls


class Boolean(schema.Boolean, _BaseElement):
    pass


class List(schema.List, _BaseElement):
    pass


class Dict(schema.Dict, _BaseElement):
    pass


class Array(schema.Array, _BaseElement):
    pass


class String(schema.String, _BaseElement):
    pass


class Integer(schema.Integer, _BaseElement):
    pass


# TODO:mp, float are currently storage only
class Float(schema.Float, _BaseElement):
    pass


class Date(schema.Date, _BaseElement):

    term_format = '%Y%m%d'
    """Format for rendering terms for this date.
    """

    value_format = '%Y%m%d'
    """Format for rendering value for this date.
    """


class DateTime(schema.DateTime, _BaseElement):

    term_format = '%Y%m%d'
    """Format for rendering terms for this date.
    """

    value_format = '%Y%m%d%H%M%S'
    """Format for rendering value for this date.
    """


class Text(schema.String, _BaseElement):

    boolean = False
    """If True, then tell Xapian to treat this field as a boolean, not
    a probablistic term.

    Most elements default as boolean, text elements do not.
    """

    language = _use_schema
    """The two letter code of the text element's language.

    If None, no language features will be applied at indexing time.

    If not None, will be used for this element, overriding the root
    schema language.

    The default is to inherit the language of the root schema.
    """

    positions = True
    """Tells the term generator to generate positional information for
    the text, or not.

    Positional information allows for phrase searching at the expense
    of a larger database.
    """

    position_start = None
    """Tells the indexer which position to start at when indexing text.

    If None, position is automatically determined on a per-document
    basis so that multiple text fields to *not* overlap.
    """

    stem = True
    """If the language is set, stem the text.
    """

    stop = True
    """If the language is set, remove stopwords from text.
    """

    spelling = True
    """True if the text should be added to the db as spelling correction targets.
    """

    string_prefix = None
    """If not None, defines the prefix that should be used for string
    indexing the text's value.

    This is useful to provide a different prefix than the name of the
    element, so that both the atomic value of the element and the
    stemmed textual value of the element can be indexed.
    """

    string = False
    """If true, also treat this field like a string.

    The elements value can also be indexed atomically with not
    language features applied.  Useful in conjunction with text_prefix
    to provide two prefixes for the text and atomic terms.
    """

    translit = 'translit/long'
    """Translit codec used to transform text terms into accent
    normalized forms.
    """


class _BaseRange(schema.Compound, _BaseElement):
    child_cls = None
    
    def __compound_init__(cls):
        cls.field_schema = [
            cls.child_cls.using(name='low',
                                optional=cls.optional),
            cls.child_cls.using(name='high',
                                optional=cls.optional),
            ]

    def compose(self):
        """Emits a tuple of low and high integers."""
        numbers = (self['low'].value, self['high'].value)
        display = ' to '.join(str(n) for n in numbers)
        return display, numbers

    def explode(self, value):
        """Consumes a sequence of low and high integers."""
        self['low'] = value[0]
        self['high'] = value[1]


class NumericRange(_BaseRange):

    child_cls = Integer

    step = 1
    """The step size for the numeric range.  Default is 1.
    """


class Location(schema.Compound, _BaseElement):
    """ Compound location is a 2-tuple of lat/lon coordinates.
    """

    radians = False
    """Presume elements coordinates are in radians.
    """
    
    child_cls = Float
    
    def __compound_init__(cls):
        cls.field_schema = [
            cls.child_cls.using(name='lat',
                                optional=cls.optional),
            cls.child_cls.using(name='lon',
                                optional=cls.optional),
            ]

    def compose(self):
        numbers = (self['lat'].value, self['lon'].value)
        display = ' by '.join(str(n) for n in numbers)
        return numbers, display

    def explode(self, value):
        self['lat'] = value[0]
        self['lon'] = value[1]

    def hash(self, radians=False):
        lat = self['lat'].value
        lon = self['lon'].value
        return geoprint.encode(lat, lon, radians=radians)



# -*- coding: utf-8 -*-

stopwords = """
 | A French stop word list. Comments begin with vertical bar. Each stop
 | word is at the start of a line.

au             |  a + le
aux            |  a + les
avec           |  with
ce             |  this
ces            |  these
dans           |  with
de             |  of
des            |  de + les
du             |  de + le
elle           |  she
en             |  `of them' etc
et             |  and
eux            |  them
il             |  he
je             |  I
la             |  the
le             |  the
leur           |  their
lui            |  him
ma             |  my (fem)
mais           |  but
me             |  me
même           |  same; as in moi-même (myself) etc
mes            |  me (pl)
moi            |  me
mon            |  my (masc)
ne             |  not
nos            |  our (pl)
notre          |  our
nous           |  we
on             |  one
ou             |  where
par            |  by
pas            |  not
pour           |  for
qu             |  que before vowel
que            |  that
qui            |  who
sa             |  his, her (fem)
se             |  oneself
ses            |  his (pl)
son            |  his, her (masc)
sur            |  on
ta             |  thy (fem)
te             |  thee
tes            |  thy (pl)
toi            |  thee
ton            |  thy (masc)
tu             |  thou
un             |  a
une            |  a
vos            |  your (pl)
votre          |  your
vous           |  you

               |  single letter forms

c              |  c'
d              |  d'
j              |  j'
l              |  l'
à              |  to, at
m              |  m'
n              |  n'
s              |  s'
t              |  t'
y              |  there

               | forms of être (not including the infinitive):
été
étée
étées
étés
étant
étante
étants
étantes
suis
es
est
sommes
êtes
sont
serai
seras
sera
serons
serez
seront
serais
serait
serions
seriez
seraient
étais
était
étions
étiez
étaient
fus
fut
fûmes
fûtes
furent
sois
soit
soyons
soyez
soient
fusse
fusses
fût
fussions
fussiez
fussent

               | forms of avoir (not including the infinitive):
ayant
ayante
ayantes
ayants
eu
eue
eues
eus
ai
as
avons
avez
ont
aurai
auras
aura
aurons
aurez
auront
aurais
aurait
aurions
auriez
auraient
avais
avait
avions
aviez
avaient
eut
eûmes
eûtes
eurent
aie
aies
ait
ayons
ayez
aient
eusse
eusses
eût
eussions
eussiez
eussent
"""



from __future__ import absolute_import

import time
import string
import logging
from collections import namedtuple
from contextlib import contextmanager

import xapian

from operator import itemgetter
from functools import partial
from json import loads

from xapian import Query, QueryParser, DocNotFoundError

from . import snowball
from .exc import ValidationError, PrefixError
from .tools import LRUDict


RETRY_LIMIT = 5
RETRY_BACKOFF_FACTOR = 0.2

logger = logging.getLogger(__name__)

XAPIAN_VERSION = namedtuple('XAPIAN_VERSION', ['major', 'minor', 'revision'])

xapian_version = XAPIAN_VERSION(xapian.major_version(),
                                  xapian.minor_version(),
                                  xapian.revision())

default_parser_flags = (QueryParser.FLAG_PHRASE |
                        QueryParser.FLAG_BOOLEAN |
                        QueryParser.FLAG_LOVEHATE |
                        QueryParser.FLAG_SPELLING_CORRECTION |
                        QueryParser.FLAG_BOOLEAN_ANY_CASE |
                        QueryParser.FLAG_WILDCARD)


def _schema_name(schema):
    return "%s.%s" % (schema.__module__, schema.__name__)


def defaults(head, limit=0, mlimit=0, klimit=1.0, kmlimit=1.0):
    return (head, limit, mlimit, klimit, kmlimit)


def _lookup_schema(name):
    modname, expr = name.rsplit('.', 1)
    local_name = modname.split('.')[-1]
    mod = __import__(modname, {}, {}, local_name)
    return eval(expr, mod.__dict__)


def _prefix(name):
    return (u'X%s:' % name.upper()).encode('utf-8')


def to_term(value, prefix=None):
    value = value.lower()
    return _prefix(prefix) + value if prefix else value


class Record(object):
    """Nice attribute-accessable record for a search result."""

    def __init__(self, schema, document, percent, rank, weight):
        self._xodb_schema = schema
        self._xodb_document = document
        self._xodb_percent = percent
        self._xodb_rank = rank
        self._xodb_weight = weight

    def __getattr__(self, name):
        try:
            return self._xodb_schema[name].value
        except KeyError:
            raise AttributeError(name)

    def __repr__(self):
        return repr(self._xodb_schema)


def record_factory(database, data, doc, percent, rank, weight):
    return Record(data, doc, percent, rank, weight)


class LanguageDecider(xapian.ExpandDecider):
    """
    A Xapian ExpandDecider that decide which terms to keep and which
    to discard when expanding a query using the "suggest" syntax.  As
    a place to start, we throw out:

      - Terms that don't begin with an uppercase letter or digit.
        This filters prefixed terms and stemmed forms.

      - Terms shorter than min_length chars, which are likely irrelevant

      - Stopwords for the given language.  Default is english, pass
        None for the language argument if no stopping is desired.
    """

    min_length = 5
    nostart = unicode(string.uppercase + string.digits)

    def __init__(self, language="en", filter=None, stems=None):
        super(LanguageDecider, self).__init__()
        if language in snowball.stoppers:
            self.stopper = snowball.stoppers[language]
            self.stemmer = xapian.Stem(language)
        else:
            self.stopper = lambda(term): False
            self.stemmer = xapian.Stem("none")
        self.stems = tuple(self.stemmer(t) for t in stems) if stems else ()

    def __call__(self, term):
        term = term.decode("utf-8")
        if (term[0] in self.nostart or
            len(term) < self.min_length or
            self.stopper(term) or
            '_' in term or
            self.stemmer(term) in self.stems):
            return False
        return True


class PrefixDecider(xapian.ExpandDecider):
    """Expand decider to only match terms that begin with a prefix."""

    __slots__ = ['prefix']

    def __init__(self, prefix):
        super(PrefixDecider, self).__init__()
        self.prefix = (u'X%s:' % prefix.upper()).encode('utf-8')

    def __call__(self, term):
        return term.startswith(self.prefix)


class MultipleValueRangeProcessor(xapian.ValueRangeProcessor):
    """Value range processor for multiple prefixes.

    :param map: a dict of prefix to value number pairs.

    :param serializer: optional callable to serialize the range arguments into
                       the same form as the corresponding values are stored.
                       Typically xapian.sortable_serialise for floats,
                       str.lower for strings.

    :param end_serializer: optional callable to serialize the end argument
                           of the range into the same form as the values.
                           Defaults to the value of *serializer*.

    """

    def __init__(self, map, serializer=None, end_serializer=None):
        self.map = map
        self.begin_serializer = serializer or (lambda x: x)
        self.end_serializer = end_serializer or self.begin_serializer
        xapian.ValueRangeProcessor.__init__(self)

    def __call__(self, begin, end):
        for prefix, value in self.map.items():
            if begin.startswith(prefix + ':'):
                return (value,
                        self.begin_serializer(begin[len(prefix) + 1:]),
                        self.end_serializer(end))
        return (xapian.BAD_VALUENO, begin, end)


class Database(object):
    """An xodb database.

    :param db_or_path: A path to file, or a xapian.Database object
    that backs this Database instance.

    :param writable: Open database in writable mode.

    :param overwrite: If writable is True, overwrite the existing
    database with a new one.

    :param spelling: If True, write spelling correction data to the
    database.

    :param replicated: If True, the database is opened read-only in
    replication mode (reopen() is never called, see xapian ticket
    #434)

    """

    record_factory = record_factory

    relevance_prefix = "_XODB_RP_"
    boolean_prefix = "_XODB_BP_"
    value_prefix = "_XODB_VALUE_"
    value_sort_prefix = "_XODB_VALUESORT_"
    value_count_name = "_XODB_COUNT_"
    backend = None
    _metadata_keyset = None
    query_cache_limit = 1024

    @contextmanager
    def transaction(self):
        self.begin()
        try:
            yield self
            self.commit()
        except:
            self.cancel()
            raise

    def retry_if_modified(self, operation, limit=RETRY_LIMIT, refresh=True):
        tries = 0
        while True:
            try:
                return operation()
            except xapian.DatabaseError, e:
                if tries >= limit:
                    logger.warning(
                        '%s after %s retries, failing.', e, tries)
                    raise
                logger.info('%s: after %s retries, retrying', e, tries)
                time.sleep(tries * RETRY_BACKOFF_FACTOR)
                self.reopen(refresh_if_needed=refresh)
                tries += 1

    def __init__(self, path,
                 writable=True,
                 overwrite=False,
                 spelling=True,
                 replicated=False,
                 inmem=False):
        self.db_path = path
        self._writable = writable
        self._overwrite = overwrite
        self.spelling = spelling
        self.replicated = replicated
        self.type_map = {}
        self.parsers_by_language = {}
        self.relevance_prefixes = {}
        self.boolean_prefixes = {}
        self.values = {}
        self.value_sorts = {}
        self.query_cache = LRUDict(limit=self.query_cache_limit)
        self.inmem = inmem
        self._value_count = 0

        if isinstance(path, basestring):
            if writable:
                if replicated:
                    raise TypeError("replication cannot be used with "
                                    "writable databases")
                if overwrite:
                    flags = xapian.DB_CREATE_OR_OVERWRITE
                else:
                    flags = xapian.DB_CREATE_OR_OPEN
                self.backend = xapian.WritableDatabase(path, flags)
            else:
                self.backend = xapian.Database(path)
        elif isinstance(path, xapian.Database):
            if replicated:
                raise TypeError("replication can only be used "
                                "if a database path is provided.")
            self.backend = self.db_path

        self.reopen()

    def get_query_parser(self, language, default_op, check_cache=True, 
                         retry_limit=RETRY_LIMIT):
        qp = None
        if check_cache and not self.inmem:
            qp = self.parsers_by_language.get(language)
            if qp is not None:
                return qp
        def prepare_op():
            return self.prepare_query_parser(language, default_op)
        qp = self.retry_if_modified(prepare_op, retry_limit)
        if check_cache and not self.inmem:
            self.parsers_by_language[language] = qp
        return qp

    def close(self):
        self.backend.close()

    def flush(self):
        self.backend.flush()

    def map(self, otype, schema):
        """Map a type to a schema."""
        self.type_map[otype] = schema

    def schema_for(self, otype):
        """Get the schema for a given type, or one of its
        superclasses."""
        for base in otype.__mro__:
            if base in self.type_map:
                return self.type_map[base]
        raise TypeError("No schema defined for %s" % repr(otype))

    def _get_value_count(self):
        if self.inmem:
            return self._value_count
        return int(self.backend.get_metadata(self.value_count_name))

    def _set_value_count(self, count):
        if self.inmem:
            self._value_count = count
        else:
            self.backend.set_metadata(self.value_count_name, str(count))

    value_count = property(_get_value_count, _set_value_count)

    def check_prefix(self, name, boolean=False):
        prefixes = set(self.relevance_prefixes.keys())
        prefixes = prefixes.union(self.boolean_prefixes.keys())
        if name not in prefixes:
            upped = _prefix(name)
            if boolean:
                self.add_boolean_prefix(name, upped)
            else:
                self.add_prefix(name, upped)

    def add_prefix(self, key, value, ignore_duplicates=True):
        """Add a prefix mapping to the database.
        """
        # FIXME:
        # Why is it only checking boolean prefixes and then adding to relevance?
        if key in self.boolean_prefixes:
            if ignore_duplicates:
                logger.warning('Conflicting relevance prefix %s', key)
                return
            else:
                raise PrefixError('Conflicting relevance prefix %s', key)
        self.relevance_prefixes[key] = value
        self.backend.set_metadata(self.relevance_prefix + key, value)

    def add_boolean_prefix(self, key, value):
        """Add a boolean prefix mapping to the database.
        """
        self.boolean_prefixes[key] = value
        self.backend.set_metadata(self.boolean_prefix + key, value)

    def allocate_value_index(self, name):
        """Default implementation of value index number allocation.

        Implement your own using shared something to keep several
        databases in sync. i.e. for indexing in parallel.
        """
        value_count = self.value_count + 1
        self.value_count = value_count
        return value_count

    def add_value(self, name, sort=None):
        """Add a value mapping to the database.
        """
        if name in self.values:
            return self.values[name]
        value_index = self.allocate_value_index(name)
        self.values[name] = value_index
        self.backend.set_metadata(self.value_prefix + name, str(value_index))
        if sort:
            self.value_sorts[name] = sort
            self.backend.set_metadata(self.value_sort_prefix + name, sort)
        return value_index

    def __len__(self):
        """ Return the number of documents in this database. """
        self.reopen()
        return self.backend.get_doccount()

    def allterms(self, prefix="", retry_limit=RETRY_LIMIT):
        self.reopen()
        seen = set()
        tries = 0
        # we can't use retry_if_modified because this
        # is not an atomic operation that returns one result
        while True:
            try:
                for t in self.backend.allterms(prefix):
                    term = t.term
                    if term in seen:
                        continue
                    seen.add(term)
                    yield term
                break
            except xapian.DatabaseError:
                if tries > retry_limit:
                    logger.warning(
                        'allterms() failed after %s retries.',
                        tries)
                    raise
                logger.info('Retrying allterms() operation.')
                time.sleep(tries * .1)
                self.reopen()
                tries += 1

    def get(self, docid, default=None):
        """ Get a document with the given docid, or the default value if
        no such document exists.
        """
        try:
            return self[docid]
        except DocNotFoundError:
            return default

    def __getitem__(self, docid):
        return self.backend.get_document(docid)

    def __delitem__(self, docid):
        self.backend.delete_document(docid)

    def __setitem__(self, docid, document):
        doc = self.get(docid)
        if doc is None:
            self.backend.add_document(document)
        else:
            self.backend.replace_document(docid, document)

    def __contains__(self, docid):
        return True if self.get(docid) else False

    def add(self, *objs, **kw):
        """Add an object to the database by transforming it into a
        xapian document.  It's type or one of its base types must be
        mapped to a schema before an object can be added here.

        :params objs: One or more mapped objects to add.

        :param schema_type: Specify the schema to be used. (optional)

        :param validate: Validated the schema before the object is
        added.  Default: True

        Returns a list of xapan documents that were added to the
        database.
        """
        assert self._writable, "Database is not writable"
        added = []
        validate = kw.pop('validate', True)
        schema_type = kw.pop('schema_type', None)
        for obj in objs:
            if isinstance(obj, xapian.Document):
                self.backend.add_document(obj)
                added.append(obj)
                continue
            doc = self.to_document(obj, schema_type=schema_type,
                                   validate=validate)
            self.backend.add_document(doc)
            added.append(doc)
        return added

    def replace(self, obj, docid, **kw):
        """Add or replace an object in the database with a specified
        document id by transforming it into a xapian document.
        It's type or one of its base types must be
        mapped to a schema before an object can be added here.

        :param obj: The mapped objects to add.

        :param docid:  The document id to user for the mapped object.

        :param schema_type: Specify the schema to be used. (optional)

        :param validate: Validated the schema before the object is
        added.  Default: True

        Returns the xapan document that was added to the
        database.
        """
        assert self._writable, "Database is not writable"
        validate = kw.pop('validate', True)
        schema_type = kw.pop('schema_type', None)
        if not isinstance(obj, xapian.Document):
            doc = self.to_document(obj, schema_type=schema_type,
                                   validate=validate)
        else:
            doc = obj
        self.backend.replace_document(int(docid), doc)
        return doc

    def to_schema(self, obj, validate=True, schema_type=None):
        """
        Turn an object into an schema instance which is fully
        populated with data from the object.  Optionally validate.
        """
        if not schema_type:
            if hasattr(obj, '__xodb_schema__'):
                schema_type = obj.__xodb_schema__
            else:
                schema_type = self.schema_for(type(obj))
        schema = schema_type.from_defaults()
        schema.update_by_object(obj)

        if validate and not schema.validate():
            invalid = []
            for child in schema.all_children:
                if not child.valid:
                    invalid.append(child)
            raise ValidationError("Elements of %s did not validate %s:" %
                                  (schema.__class__.__name__,
                                   list((c.name, c.value)
                                        for c in invalid)))
        return schema

    def to_document(self, obj, validate=True, schema_type=None):
        """
        Convienient wrapper that does the object->schema->document
        transformation.
        """
        schema = self.to_schema(obj, validate, schema_type)
        return self.doc_from_dict(schema.memo.dict)

    def doc_from_dict(self, data):
        """Take an intermediate representation of a document (a
        "memo") and turn it into a xapian document.  The document is
        returned and not added to the database.
        """
        doc = xapian.Document()

        for term_tup in data.get('terms', ()):
            term = None
            typ = None
            wdfinc = 1
            if len(term_tup) > 2:
                term, typ, wdfinc = term_tup
            else:
                term, typ = term_tup
            if ':' in term:
                prefix, _, value = term.partition(':')
                # FIXME: check_prefix has a second param called boolean=False
                # Is passing the typ in the right thing to do?
                self.check_prefix(prefix, typ)
                term = 'X%s:%s' % (prefix.upper(), value)
            doc.add_term(term, wdfinc)
        for post in data.get('posts', ()):
            pass

        all_start_pos = 0
        for text_dict in data.get('texts', ()):
            lang = text_dict.get('lang')
            el_start_pos = text_dict.get('position_start', None)

            tg = xapian.TermGenerator()
            tg.set_database(self.backend)
            tg.set_document(doc)

            # if the element specifies no start position, set the
            # starting position where the last element left off
            if el_start_pos is not None:
                tg.set_termpos(el_start_pos)
            else:
                tg.set_termpos(all_start_pos)

            spelling = text_dict.get('spell', True)
            if spelling:
                try:
                    # hack to workaround missing spelling for inmem backends
                    self.backend.add_spelling('food')
                    tg.set_flags(xapian.TermGenerator.FLAG_SPELLING)
                except:
                    pass  # noop for backends that don't support spelling
            if lang in snowball.stoppers:
                tg.set_stemmer(xapian.Stem(lang))
                tg.set_stopper(snowball.stoppers[lang])
            if text_dict.get('post', True):
                index_text = tg.index_text
            else:
                index_text = tg.index_text_without_positions
            text = text_dict.get('text')
            if text:
                wdf_inc = text_dict.get('wdf_inc', 1)
                prefix = text_dict.get('prefix')
                if prefix:
                    self.check_prefix(prefix)
                    prefix = 'X%s:' % prefix.upper()
                    index_text(text, wdf_inc, prefix)
                else:
                    index_text(text, wdf_inc)
            # if the element specified no start position,
            # update the all-document position
            if el_start_pos is None:
                all_start_pos = tg.get_termpos()

        for val_tuple in data.get('values', ()):
            name, value, typ = val_tuple
            if name in self.values:
                valno = self.values[name]
            else:
                valno = self.add_value(name, typ)
            if typ in ('integer',):
                value = xapian.sortable_serialise(value)
            doc.add_value(valno, value)

        data = data.get('data')
        if data:
            doc.set_data(data)
        return doc

    def reopen(self, retry_limit=RETRY_LIMIT, refresh_if_needed=True):
        """
        Reopen the database.  Called before most query methods.  If
        replication is used, the db is closed and reopened.
        """
        if not self.replicated:
            self.backend.reopen()
        else:
            # replication does not support the reopen() method, so the
            # db must be explicitly closed and reopened.
            assert self.db_path, ("Must provide a db path when "
                                  "using replication.")
            self.close()
            self.backend = xapian.Database(self.db_path)
            # reset cached parsers to new database object
            for parser in self.parsers_by_language.itervalues():
                parser.set_database(self.backend)

        if refresh_if_needed and self.is_metadata_changed:
            self.meta_refresh()

    def begin(self):
        if self._writable:
            self.reopen()
            try:
                self.backend.begin_transaction()
            except Exception:
                pass  # noop for backends that don't support transactions

    def cancel(self):
        if self._writable:
            try:
                self.backend.cancel_transaction()
            except Exception:
                pass

    def commit(self):
        if self._writable:
            try:
                self.backend.commit_transaction()
            except Exception:
                pass

    def prepare_query_parser(self, language=None,
                             default_op=Query.OP_AND):
        """
        Setup a query parser with the current known prefixes and values.
        """
        qp = QueryParser()
        qp.set_database(self.backend)
        qp.set_default_op(default_op)
        
        if self.boolean_prefixes:
            for key, value in self.boolean_prefixes.items():
                qp.add_boolean_prefix(key, value)
        if self.relevance_prefixes:
            for key, value in self.relevance_prefixes.items():
                if key not in self.boolean_prefixes:
                    qp.add_prefix(key, value)
                else:
                    logger.warning(
                        'Duplicate relevance prefix %s conflicts with boolean',
                        key)
        if self.value_sorts:
            # First add numeric values ranges
            qp.add_valuerangeprocessor(MultipleValueRangeProcessor(
                dict(((k, self.values[k])
                      for k, v in self.value_sorts.items() if v == 'integer')),
                lambda s: xapian.sortable_serialise(float(s)),
            ))
            # Then string and date
            qp.add_valuerangeprocessor(MultipleValueRangeProcessor(
                dict(((k, self.values[k])
                      for k, v in self.value_sorts.items() if v in
                        ('string', 'date'))),
            ))
            # Serialize date range queries so that they are inclusive.
            # This allows datetime value range queries to be treated
            # as [begin,end] rather than [begin,end) as is the default
            # without these serializers when then datetime range arguments
            # are not fully qualified.
            qp.add_valuerangeprocessor(MultipleValueRangeProcessor(
                dict(((k, self.values[k])
                      for k, v in self.value_sorts.items() if v == 'datetime')),
                serializer = lambda x: x + '0'*(14-len(x)),
                end_serializer = lambda x: x + '9'*(14-len(x))
            ))
        if language in snowball.stoppers:
            qp.set_stemmer(xapian.Stem(language))
            qp.set_stopper(snowball.stoppers[language])
            qp.set_stemming_strategy(QueryParser.STEM_SOME)
        return qp

    @property
    def is_metadata_changed(self):
        return self._metadata_keyset != self._get_metadata_keyset()

    def _get_metadata_keyset(self, retry_limit=RETRY_LIMIT):
        if self.inmem:
            return False
        op = lambda: set(self.backend.metadata_keys())
        # don't recurse into refresh here, just reopen and retry
        return self.retry_if_modified(op, retry_limit, False)

    def meta_refresh(self, retry_limit=RETRY_LIMIT):
        if not self.inmem:
            self.parsers_by_language = {}
            self.relevance_prefixes = {}
            self.boolean_prefixes = {}
            self.values = {}
            self.value_sorts = {}
            self.query_cache = LRUDict(limit=self.query_cache_limit)

            self._metadata_keyset = self._get_metadata_keyset()
            for k in self._metadata_keyset:
                op = lambda: self.backend.get_metadata(k)
                val = self.retry_if_modified(op, retry_limit, False)
                if k.startswith(self.relevance_prefix):
                    prefix = k[len(self.relevance_prefix):]
                    self.relevance_prefixes[prefix] = val
                elif k.startswith(self.boolean_prefix):
                    prefix = k[len(self.boolean_prefix):]
                    self.boolean_prefixes[prefix] = val
                elif k.startswith(self.value_prefix):
                    value = k[len(self.value_prefix):]
                    self.values[value] = int(val)
                elif k.startswith(self.value_sort_prefix):
                    value = k[len(self.value_sort_prefix):]
                    self.value_sorts[value] = val
            try:
                # hit the property to refresh this value
                count = self.value_count
                count = count  # pyflakes
            except ValueError:
                if self._writable:
                    self.value_count = 0

    def querify(self, query,
                language=None,
                translit=None,
                default_op=Query.OP_AND,
                parser_flags=default_parser_flags,
                retry_limit=RETRY_LIMIT):
        """Return a query object, constructed from a string, query
        object, or mixed sequence of any number of strings, query
        objects, or subsequences of the same form.

        If 'query' is a xapian query object, it is returned unchanged.
        If it's a string, it is parsed with xapian's query parser and
        returned.  If it is a sequence, a new query is constructed
        with the 'default_op' operator and the sequence is iterated
        into the new query, recursively querify on each item of the
        sequence.
        """
        if isinstance(query, Query):
            return query
        if isinstance(query, basestring):
            if query == "":
                return Query("")
            else:
                cache_key = (query, language, translit, default_op, parser_flags)
                if cache_key in self.query_cache:
                    return self.query_cache[cache_key]
                query = query.lower()
                if translit:
                    query = query.encode(translit)

                qp = self.get_query_parser(language, default_op, 
                                           retry_limit=retry_limit)
                def query_op():
                    return qp.parse_query(query, parser_flags)
                result = self.retry_if_modified(query_op, retry_limit)
                if not self.inmem:
                    self.query_cache[cache_key] = result
                return result
        else:
            return reduce(partial(Query, default_op),
                          (self.querify(q,
                                        language,
                                        translit,
                                        default_op,
                                        parser_flags) for q in query))

    def query(self, query,
              offset=0,
              limit=0,
              order=None,
              reverse=False,
              language=None,
              check=0,
              translit=None,
              match_decider=None,
              match_spy=None,
              document=False,
              echo=False,
              parser_flags=default_parser_flags,
              default_op=Query.OP_AND,
              retry_limit=RETRY_LIMIT):
        """
        Query the database with the provided string or xapian Query
        object.  A string is passed into xapians QueryParser first to
        generate a Query object.
        """
        self.reopen()

        enq = xapian.Enquire(self.backend)
        if echo:
            print "Parsing query."
        query = self.querify(query, language, translit,
                             default_op, parser_flags)
        if echo:
            print "Done parsing query: %s" % str(query)
        enq.set_query(query)

        limit = limit or self.backend.get_doccount()

        if echo:
            start = time.time()
            print "Fetching mset..."

        # convoluted logic here is to retry queries that die in the
        # middle of result iteration because the db was closed (due to
        # replication).

        tries = 0
        seen = set()
        while True:
            try:
                # _build_mset may retry internally on DatabaseError
                mset = self._build_mset(enq, offset, limit, order, reverse,
                                        check, match_decider, match_spy,
                                        retry_limit=retry_limit)
                if echo:
                    print "Fetched mset in %s" % str(time.time() - start)

                for record in mset:
                    doc = record.document
                    docid = doc.get_docid()
                    if docid in seen:
                        continue
                    if document:
                        seen.add(docid)
                        yield doc
                    else:
                        # retry getting the actual document data
                        op = lambda: self.backend.get_document(docid).get_data()
                        try:
                            data = self.retry_if_modified(op, retry_limit)
                        except DocNotFoundError:
                            logger.warning(
                                "Document %d is gone, skipping.", docid)
                            continue
                        typ, data = loads(data)
                        seen.add(docid)
                        yield self.record_factory(_lookup_schema(typ).from_flat(data),
                                                  doc,
                                                  record.percent,
                                                  record.rank,
                                                  record.weight,
                                                  )
                # no errors exhuasting the set? break out and we're done
                break
            except xapian.DatabaseError, e:
                # an error occured, either, the db was closed, or the
                # modified error happened two frequently in the inner
                # loop, so we are going to replay the whole query
                if tries > retry_limit:
                    logger.warning(
                        'Database replay failed after %s retries.',
                        tries)
                    raise
                logger.info('Replaying database query.')
                self.reopen()
                tries += 1

    def count(self,
              query="",
              language=None,
              echo=False,
              translit=None,
              parser_flags=default_parser_flags,
              default_op=Query.OP_AND,
              retry_limit=RETRY_LIMIT):
        """
        Query the database with the provided string or xapian Query
        object.  A string is passed into xapians QueryParser first to
        generate a Query object.
        """
        self.reopen()
        query = self.querify(query, language, translit, default_op, parser_flags)
        if echo:
            print str(query)
        enq = xapian.Enquire(self.backend)
        enq.set_query(query)

        mset = self._build_mset(enq, retry_limit=retry_limit)
        return mset.size()

    def facet(self, query,
              prefix='facet',
              estimate=True,
              language=None,
              limit=0,
              mlimit=0,
              klimit=1.0,
              kmlimit=1.0,
              echo=False,
              retry_limit=RETRY_LIMIT):
        """Get facet suggestions for the query, then the query with
        each suggested facet, asking xapian for an estimated count of
        each sub-query.
        """
        if estimate:
            counter = self.estimate
        else:
            counter = self.count

        results = {}
        query = self.querify(query, language=language)

        suggestions = self.suggest(query,
                                   prefix=prefix,
                                   language=language,
                                   limit=limit,
                                   mlimit=mlimit,
                                   klimit=klimit,
                                   kmlimit=kmlimit,
                                   echo=echo,
                                   retry_limit=retry_limit,
                                   format_term=False)
        for facet in suggestions:
            q = Query(Query.OP_AND, [query, facet])
            if echo:
                print str(q)
            if prefix and facet.startswith('X%s:' % prefix.upper()):
                suffix = facet[len(prefix) + 2:]
                if ' ' in suffix or '..' in suffix:
                    facet = '%s:"%s"' % (prefix, suffix)
                else:
                    facet = '%s:%s' % (prefix, suffix)
            results[facet] = counter(q, language=language)
        return results

    def expand(self, query, expand,
               language=None,
               echo=False,
               translit=None,
               default_op=Query.OP_AND,
               parser_flags=default_parser_flags,
               retry_limit=RETRY_LIMIT):
        """
        Expand a query on a given set of prefixes.
        """

        q = self.querify(query, language, translit, default_op, parser_flags,
                         retry_limit=retry_limit)
        if echo:
            print q
        results = {}
        head, tail = expand[0], expand[1:]
        if isinstance(head, tuple):
            args = head
        else:
            args = (head,)
        head, limit, mlimit, klimit, kmlimit = defaults(*args)

        r = self.facet(
            q, prefix=head, language=language, echo=echo,
            limit=limit, mlimit=mlimit,
            klimit=klimit, kmlimit=kmlimit,
            retry_limit=retry_limit).items()

        for name, score in r:
            if tail:
                subq = self.querify([q, name], retry_limit=retry_limit)
                if echo:
                    print subq
                r = self.expand(
                    subq, tail, language, echo,
                    default_op, parser_flags,
                    retry_limit=retry_limit)
                results[(name, score)] = r
            else:
                results[(name, score)] = score
        return results

    def estimate(self, query,
                 limit=0,
                 klimit=1.0,
                 language=None,
                 translit=None,
                 default_op=Query.OP_AND,
                 parser_flags=default_parser_flags,
                 retry_limit=RETRY_LIMIT):
        """Estimate the number of documents that will be yielded with the
        given query.

        Limit tells the estimator the minimum number of documents to
        consider.  A zero limit means potentially check all documents
        in the db."""
        self.reopen()
        enq = xapian.Enquire(self.backend)

        if limit == 0:
            limit = int(self.backend.get_doccount() * klimit)

        query = self.querify(query, language, translit,
                             default_op, parser_flags,
                             retry_limit=retry_limit)

        enq.set_query(query)
        op = lambda: enq.get_mset(0, 0, limit)
        mset = self.retry_if_modified(op, retry_limit)

        return mset.get_matches_estimated()

    def term_freq(self, term):
        """
        Return a count of the number of documents indexed for a given
        term.  Useful for testing.
        """
        self.reopen()
        return self.backend.get_termfreq(term)

    def describe_query(self, query,
                       language=None,
                       default_op=Query.OP_AND,
                       retry_limit=RETRY_LIMIT):
        """
        Describe the parsed query.
        """
        qp = self.get_query_parser(language, default_op,
                                   retry_limit=RETRY_LIMIT)
        def op():
            return qp.parse_query(query, default_parser_flags) 
        return str(self.retry_if_modified(op, retry_limit))

    def spell(self, query,
              language=None,
              default_op=Query.OP_AND,
              retry_limit=RETRY_LIMIT):
        """
        Suggest a query string with corrected spelling.
        """
        self.reopen()
        qp = self.get_query_parser(language, default_op, 
                                   retry_limit=RETRY_LIMIT)
        def op():
            qp.parse_query(query, QueryParser.FLAG_SPELLING_CORRECTION)
            return qp.get_corrected_query_string().decode('utf8')
        return self.retry_if_modified(op, retry_limit)

    def suggest(self, query,
                offset=0,
                limit=0,
                moffset=0,
                mlimit=0,
                klimit=1.0,
                kmlimit=1.0,
                translit=None,
                language=None,
                prefix=None,
                decider=None,
                score=False,
                echo=False,
                default_op=Query.OP_AND,
                parser_flags=default_parser_flags,
                retry_limit=RETRY_LIMIT,
                format_term=True):
        """
        Suggest terms that would possibly yield more relevant results
        for the given query.
        """
        self.reopen()
        enq = xapian.Enquire(self.backend)

        query = self.querify(query, language, translit, default_op, parser_flags)

        if mlimit == 0:
            mlimit = int(self.backend.get_doccount() * kmlimit)

        if echo:
            print str(query)
        enq.set_query(query)

        mset = self._build_mset(enq, offset=moffset, limit=mlimit,
                                retry_limit=retry_limit)

        rset = xapian.RSet()
        for m in mset:
            rset.add_document(m.docid)

        if prefix is not None:
            decider = PrefixDecider(prefix)

        if decider is None:
            decider = LanguageDecider(language)

        if limit == 0:
            limit = int(self.backend.get_doccount() * klimit)

        if xapian_version <= (1, 2, 4):
            op = lambda: enq.get_eset(
                limit, rset,
                enq.INCLUDE_QUERY_TERMS,
                1.0, decider)
        else:
            op = lambda: enq.get_eset(
                limit, rset,
                enq.INCLUDE_QUERY_TERMS,
                1.0, decider, -3)

        eset = self.retry_if_modified(op, retry_limit)

        for item in eset.items:
            val = item[0].decode('utf8')
            if format_term and prefix and val.startswith('X%s:' % prefix.upper()):
                suffix = val[len(prefix) + 2:]
                if ' ' in suffix or '..' in suffix:
                    val = '%s:"%s"' % (prefix, suffix)
                else:
                    val = '%s:%s' % (prefix, suffix)
            if score:
                yield (val, item[1])
            else:
                yield val

    def _build_mset(self, enq,
                    offset=0,
                    limit=None,
                    order=None,
                    reverse=False,
                    check=None,
                    match_decider=None,
                    match_spy=None,
                    retry_limit=RETRY_LIMIT):
        if order is not None:
            if isinstance(order, basestring):
                try:
                    order = self.values[order]
                except KeyError:
                    raise ValueError("There is no sort name %s" % order)
            enq.set_sort_by_value(order, reverse)

        if limit is None:
            limit = self.backend.get_doccount()

        if check is None:
            check = limit + 1

        op = lambda: enq.get_mset(
                offset, limit, check, None, match_decider, match_spy)
        return self.retry_if_modified(op, retry_limit)


def jsonrpc_wrapper(f):
    def w(database, *args, **kw):
        # unstupidify the way the jsonrpc spec calls us
        if not args:
            args = []
            for k in kw:
                try:
                    i = int(k)
                except ValueError:
                    continue
                args.append((i, kw.pop(k)))
            args.sort(key=itemgetter(0))
            args = tuple(a[1] for a in args)
        return f(database, *args, **kw)
    return w


class JSONDatabase(object):
    """
    Thunk layer on top of database that returns json data.
    """

    def jsonify_record(database, record):
        return record.flatten()

    def __init__(self, database):
        database.record_factory = self.jsonify_record
        self._database = database

    @jsonrpc_wrapper
    def query(self, *args, **kw):
        return list(self._database.query(*args, **kw))

    @jsonrpc_wrapper
    def suggest(self, *args, **kw):
        return list(self._database.query(*args, **kw))

    @jsonrpc_wrapper
    def facet(self, *args, **kw):
        return self._database.facet(*args, **kw)

    @jsonrpc_wrapper
    def expand(self, *args, **kw):
        return self._database.expand(*args, **kw)

    @jsonrpc_wrapper
    def count(self, *args, **kw):
        return self._database.count(*args, **kw)



import sys

ADDRESS = 'localhost'

PORT = 2730

URI = 'http://%s:%s/' % (ADDRESS, PORT)


def make_application(database, address=ADDRESS, port=PORT):
    from xodb import JSONDatabase
    import wsgi_jsonrpc

    database = JSONDatabase(database)
    application = wsgi_jsonrpc.WSGIJSONRPCApplication(instance=database)
    return application


def run_server(database, address=ADDRESS, port=PORT,
               reloader=None, threaded=True, processes=1,
               errors_fatal=False, echo_sql=False):
    from werkzeug.serving import run_simple

    app = make_application(database, address, port)

    print >> sys.stderr, " * Listening on %s:%s." % (address, port)
    run_simple(address, port, app,
               use_reloader=reloader,
               reloader_interval=1,
               threaded=threaded,
               processes=processes,
               passthrough_errors=errors_fatal,
               )




import pyparsing

"""
The query language that won't die.

Syntax:

    Typical search engine query language, terms with boolean operators
    and parenthesized grouping:

      (term AND (term OR term OR ...) AND NOT term ...)

    In it's simplest case, xaql searches for a list of terms:

      term term term ...

    This expands to '(term AND term AND term AND ...)'

    Terms can be prefixed.  The prefix and the value are separated by
    colons.  Values that contain spaces must be double quoted.

      term color:blue name:"My Lagoon"


Functions:

    Functions provide features that take query input from the user and
    do some transformation on the query itself.  Functions begin with
    a star, then a name, then a pair of parenthesis that contain the
    query input.  The syntax of the input is up to the function:

    $xp(...) -- Pass the input string directly into the Xapian
     query parser.

    $rel(...)

    $now


Prefix Modifiers:

  This are pre-prefixes that transform the following term.

    published-before:now
    

"""






MAX_TERM_LEN = 240 # compile time xapian limit

from .exc import InvalidTermError


class Memo(object):
    """ A lightweight, picklable container for document data.  Contain
    all the data elements necessary to construct a xapian document."""

    def __init__(self):
        self.lang = None
        self.terms = []    # (term, type) | (term, type, wdfinc)
        self.posts = []    # (term, position) | (term, position, wdfinc)
        self.texts = []    # {text=u"", lang=None, stem=True, stop=True,
                           #  posit=True, post_start=None}
        self.values = []   # (name, value, type)
        self.data = None   # bytestring or None

    @classmethod
    def from_dict(cls, data):
        """ Construct a Memo from a data dictionary. """
        m = cls()
        m.lang = data.get('lang')
        m.terms = data.get('terms', [])
        m.posts = data.get('posts', [])
        m.texts = data.get('texts', [])
        m.values = data.get('values', [])
        m.data = data.get('data')
        return m

    @property
    def dict(self):
        """ Returns a dictionary representation of the document. """
        return dict(
            lang=self.lang,
            terms=self.terms,
            posts=self.posts,
            texts=self.texts,
            values=self.values,
            data=self.data)

    def add_term(self, term, boolean=False, wdfinc=None):
        if len(term) > MAX_TERM_LEN:
            raise InvalidTermError("The term %s is too long" % term)
        typ = 'b' if boolean else 'r'
        term = (term, typ, wdfinc) if wdfinc else (term, typ)
        self.terms.append(term)

    def add_post(self, term, position, wdfinc=None):
        term = (term, position, wdfinc) if wdfinc else (term, position)
        self.posts.append(term)

    def add_text(self, text, prefix=None, lang=None,
                 post=True, stem=True, stop=True, spell=True,
                 wdf_inc=None, post_start=None):
        self.texts.append(dict(text=text,
                               prefix=prefix,
                               lang=lang,
                               post=post,
                               stem=stem,
                               stop=stop,
                               spell=spell,
                               wdf_inc=wdf_inc,
                               post_start=post_start,
                               ))

    def add_value(self, name, value, type):
        self.values.append((name, value, type))

    def set_data(self, data):
        self.data = data

    def set_lang(self, lang):
        self.lang = lang





class XODBError(Exception):
    pass


class InvalidTermError(XODBError):
    pass


class AttributeRequired(XODBError):
    pass


class AttributeTypeRequired(XODBError):
    pass


class ValidationError(XODBError):
    pass


class PrefixError(XODBError):
    pass



# OrderedDict is cargoed from SQLAlchemy,
# LRUDict is cargoed from Idealist.org


class OrderedDict(dict):
    """A dict that returns keys/values/items in the order they were added."""

    def __init__(self, ____sequence=None, **kwargs):
        self._list = []
        if ____sequence is None:
            if kwargs:
                self.update(**kwargs)
        else:
            self.update(____sequence, **kwargs)

    def clear(self):
        self._list = []
        dict.clear(self)

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return OrderedDict(self)

    def sort(self, *arg, **kw):
        self._list.sort(*arg, **kw)

    def update(self, ____sequence=None, **kwargs):
        if ____sequence is not None:
            if hasattr(____sequence, 'keys'):
                for key in ____sequence.keys():
                    self.__setitem__(key, ____sequence[key])
            else:
                for key, value in ____sequence:
                    self[key] = value
        if kwargs:
            self.update(kwargs)

    def setdefault(self, key, value):
        if key not in self:
            self.__setitem__(key, value)
            return value
        else:
            return self.__getitem__(key)

    def __iter__(self):
        return iter(self._list)

    def values(self):
        return [self[key] for key in self._list]

    def itervalues(self):
        return iter(self.values())

    def keys(self):
        return list(self._list)

    def iterkeys(self):
        return iter(self.keys())

    def items(self):
        return [(key, self[key]) for key in self.keys()]

    def iteritems(self):
        return iter(self.items())

    def __setitem__(self, key, object):
        if key not in self:
            try:
                self._list.append(key)
            except AttributeError:
                # work around Python pickle loads() with 
                # dict subclass (seems to ignore __setstate__?)
                self._list = [key]
        dict.__setitem__(self, key, object)

    def __delitem__(self, key):
        dict.__delitem__(self, key)
        self._list.remove(key)

    def pop(self, key, *default):
        present = key in self
        value = dict.pop(self, key, *default)
        if present:
            self._list.remove(key)
        return value

    def popitem(self):
        item = dict.popitem(self)
        self._list.remove(item[0])
        return item


class LRUDict(OrderedDict):
    """
    A simple data structure to provide item ordering based on usage for a
    limited dictionary size. The last item in the dictionary will always be
    considered the most recently used. By the same logic, the item at the front
    will be the least recently used. If the dictionary is filled beyond the
    limit, items at the front of the list will be discarded.
    """

    def __init__(self, items=(), limit=1000, on_cache_miss=None):
        self.limit = limit
        super(LRUDict, self).__init__(items)

    def __setitem__(self, key, value):
        if len(self) >= self.limit:
            self.popitem(False)
        super(LRUDict, self).__setitem__(key, value)

    def __getitem__(self, key):
        item = dict.pop(self, key)
        self[key] = item
        return item

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default



from __future__ import absolute_import

import tempfile
import xapian

from . tools import geoprint

from . database import (
    Database,
    JSONDatabase,
    LanguageDecider,
    MultipleValueRangeProcessor,
    )

from . elements import (
    Array,
    Date,
    DateTime,
    Dict,
    Integer,
    List,
    Location,
    NumericRange,
    Schema,
    String,
    Text,
    )

__all__ = [
    'Array',
    'Database',
    'Date',
    'DateTime',
    'Dict',
    'Integer',
    'JSONDatabase',
    'LanguageDecider',
    'List',
    'Location',
    'MultipleValueRangeProcessor',
    'NumericRange',
    'Schema',
    'String',
    'Text',
    'geoprint',
    'inmemory',
    'open',
    'temp',
    ]


def open(path_or_db,
         writable=True,
         overwrite=False,
         spelling=True,
         replicated=False,
         inmem=False):
    """Return an xodb database with the given path or xapian database object.

    :param path_or_db: A path to a database file or a pre-existing
    xapian database object.

    :param writable: Open database in writable mode.

    :param overwrite: If writable is True, overwrite the existing
    database with a new one.

    :param spelling: If True, write spelling correction data to the
    database.
    """
    return Database(path_or_db,
                    writable=writable,
                    overwrite=overwrite,
                    spelling=spelling,
                    replicated=replicated,
                    inmem=inmem)


def temp(spelling=True):
    """Returns an xodb database backed by a teporary directory.  You
    are responsible for cleaning up the directory.

    :param spelling: If True, write spelling correction data to database.
    """
    return open(tempfile.mkdtemp(), spelling=spelling)


def inmemory():
    """Returns an xodb database backed by an in-memory xapian
    database.  Does not support spelling correction.
    """
    return open(xapian.inmemory_open(), spelling=False, inmem=True)



from __future__ import absolute_import

from cStringIO import StringIO

from xapian import SimpleStopper
from . import english, spanish, french, german, italian, russian

def build_stopwords(language, encoding="utf8"):
    file = StringIO(language.stopwords)
    stopwords = []
    for line in file.readlines():
        word = unicode(line.strip().split("|")[0].strip(), encoding)
        if word:
            stopwords.append(word)
    return stopwords

stopwords = {
    "en" : build_stopwords(english, encoding="latin-1"),
    "es" : build_stopwords(spanish, encoding="latin-1"),
    "ru" : build_stopwords(russian, encoding="koi8_r"),
    "fr" : build_stopwords(french, encoding="latin-1"),
    "de" : build_stopwords(german, encoding="latin-1"),
    "it" : build_stopwords(italian, encoding="latin-1"),
}

stoppers = {}

for code in stopwords:
    stopper = SimpleStopper()
    for word in stopwords[code]:
        stopper.add(word)
    stoppers[code] = stopper








#!/usr/bin/env python
"""
====
xodb
====

Development Version
-------------------

The hg `xodb tip`_ can be installed via ``easy_install xodb==dev``.

.. _xodb tip: http://bitbucket.org/pelletier_michel/xodb/get/tip.zip#egg=xodb-dev

"""
from setuptools import setup, find_packages

setup(
  name = "xodb",
  version = "0.4.17",
  packages=find_packages(exclude=['tests.*', 'tests']),

  tests_require=['nose', 'translitcodec'],
  test_suite='nose.collector',

  author='Michel Pelletier - Action Without Borders',
  author_email='michel@idealist.org',
  description='experimental xapian object database',
  long_description=__doc__,
  license='MIT License',
  url='http://bitbucket.org/pelletier_michel/xodb/',
      install_requires=[
        'flatland',
        'translitcodec',
        ],
  classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.5',
        'Programming Language :: Python :: 2.6',
        ],
)



# -*- coding: utf-8 -*-

stopwords = """
 | A Spanish stop word list. Comments begin with vertical bar. Each stop
 | word is at the start of a line.


 | The following is a ranked list (commonest to rarest) of stopwords
 | deriving from a large sample of text.

 | Extra words have been added at the end.

de             |  from, of
la             |  the, her
que            |  who, that
el             |  the
en             |  in
y              |  and
a              |  to
los            |  the, them
del            |  de + el
se             |  himself, from him etc
las            |  the, them
por            |  for, by, etc
un             |  a
para           |  for
con            |  with
no             |  no
una            |  a
su             |  his, her
al             |  a + el
  | es         from SER
lo             |  him
como           |  how
más            |  more
pero           |  pero
sus            |  su plural
le             |  to him, her
ya             |  already
o              |  or
  | fue        from SER
este           |  this
  | ha         from HABER
sí             |  himself etc
porque         |  because
esta           |  this
  | son        from SER
entre          |  between
  | está     from ESTAR
cuando         |  when
muy            |  very
sin            |  without
sobre          |  on
  | ser        from SER
  | tiene      from TENER
también        |  also
me             |  me
hasta          |  until
hay            |  there is/are
donde          |  where
  | han        from HABER
quien          |  whom, that
  | están      from ESTAR
  | estado     from ESTAR
desde          |  from
todo           |  all
nos            |  us
durante        |  during
  | estados    from ESTAR
todos          |  all
uno            |  a
les            |  to them
ni             |  nor
contra         |  against
otros          |  other
  | fueron     from SER
ese            |  that
eso            |  that
  | había      from HABER
ante           |  before
ellos          |  they
e              |  and (variant of y)
esto           |  this
mí             |  me
antes          |  before
algunos        |  some
qué            |  what?
unos           |  a
yo             |  I
otro           |  other
otras          |  other
otra           |  other
él             |  he
tanto          |  so much, many
esa            |  that
estos          |  these
mucho          |  much, many
quienes        |  who
nada           |  nothing
muchos         |  many
cual           |  who
  | sea        from SER
poco           |  few
ella           |  she
estar          |  to be
  | haber      from HABER
estas          |  these
  | estaba     from ESTAR
  | estamos    from ESTAR
algunas        |  some
algo           |  something
nosotros       |  we

      | other forms

mi             |  me
mis            |  mi plural
tú             |  thou
te             |  thee
ti             |  thee
tu             |  thy
tus            |  tu plural
ellas          |  they
nosotras       |  we
vosostros      |  you
vosostras      |  you
os             |  you
mío            |  mine
mía            |
míos           |
mías           |
tuyo           |  thine
tuya           |
tuyos          |
tuyas          |
suyo           |  his, hers, theirs
suya           |
suyos          |
suyas          |
nuestro        |  ours
nuestra        |
nuestros       |
nuestras       |
vuestro        |  yours
vuestra        |
vuestros       |
vuestras       |
esos           |  those
esas           |  those

               | forms of estar, to be (not including the infinitive):
estoy
estás
está
estamos
estáis
están
esté
estés
estemos
estéis
estén
estaré
estarás
estará
estaremos
estaréis
estarán
estaría
estarías
estaríamos
estaríais
estarían
estaba
estabas
estábamos
estabais
estaban
estuve
estuviste
estuvo
estuvimos
estuvisteis
estuvieron
estuviera
estuvieras
estuviéramos
estuvierais
estuvieran
estuviese
estuvieses
estuviésemos
estuvieseis
estuviesen
estando
estado
estada
estados
estadas
estad

               | forms of haber, to have (not including the infinitive):
he
has
ha
hemos
habéis
han
haya
hayas
hayamos
hayáis
hayan
habré
habrás
habrá
habremos
habréis
habrán
habría
habrías
habríamos
habríais
habrían
había
habías
habíamos
habíais
habían
hube
hubiste
hubo
hubimos
hubisteis
hubieron
hubiera
hubieras
hubiéramos
hubierais
hubieran
hubiese
hubieses
hubiésemos
hubieseis
hubiesen
habiendo
habido
habida
habidos
habidas

               | forms of ser, to be (not including the infinitive):
soy
eres
es
somos
sois
son
sea
seas
seamos
seáis
sean
seré
serás
será
seremos
seréis
serán
sería
serías
seríamos
seríais
serían
era
eras
éramos
erais
eran
fui
fuiste
fue
fuimos
fuisteis
fueron
fuera
fueras
fuéramos
fuerais
fueran
fuese
fueses
fuésemos
fueseis
fuesen
sintiendo
sentido
sentida
sentidos
sentidas
siente
sentid

               | forms of tener, to have (not including the infinitive):
tengo
tienes
tiene
tenemos
tenéis
tienen
tenga
tengas
tengamos
tengáis
tengan
tendré
tendrás
tendrá
tendremos
tendréis
tendrán
tendría
tendrías
tendríamos
tendríais
tendrían
tenía
tenías
teníamos
teníais
tenían
tuve
tuviste
tuvo
tuvimos
tuvisteis
tuvieron
tuviera
tuvieras
tuviéramos
tuvierais
tuvieran
tuviese
tuvieses
tuviésemos
tuvieseis
tuviesen
teniendo
tenido
tenida
tenidos
tenidas
tened
"""



"""Geoprint is a geohash variant.

This algorithm is almost exactly identical to, and originally based
on, the Geohash module on Pypi by Leonard Norrgard, which in turn,
implements the geohash algorithm developed by Gustavo Niemeyer and
described in the wikipedia article:

http://en.wikipedia.org/wiki/Geohash

The main difference between this encoding and geohash is that it uses
a much smaller 'alphabet' of encoding characters to spell a hash,
instead of base32 encoding it uses simpler base4 encoding that uses
four possible states to encode each character instead of 32.  This
results in longer, uglier, less compressed geohashes but they are more
useful for doing proximity searching using hash prefixes.  Note that
like geohashes, there are edge cases near the equator and meridians
where points near each other have different prefixes.

All geoprints begin with either a 'w' or 'e' character to specify that
the hash belongs in the western or eastern hemisphere.  The western
hemisphere includes all longitudes less than zero down to -180.0, and
the eastern hemisphere includes all longitudes greater than or equal
to zero, up to 180.0.  Subsequent hash characters are one of g, a, t,
or c.
"""

from collections import namedtuple
from operator import neg, mod
from math import radians as rads
from math import (
    asin,
    atan2,
    cos,
    degrees,
    log10,
    pi,
    sin,
    sqrt,
    )

interval = namedtuple("interval", "min max")

alphabet = 'gatc'
decodemap = dict((k, i) for (i, k) in enumerate(alphabet))

EARTH_RADIUS = 6378100

def encode(latitude, longitude, precision=22, radians=False):
    """Encode the given latitude and longitude into a geoprint that
    contains 'precision' characters.

    If radians is True, input parameters are in radians, otherwise
    they are degrees.  Example::

    >>> c = (7.0625, -95.677068)
    >>> h = encode(*c)
    >>> h
    'watttatcttttgctacgaagt'

    >>> r = rads(c[0]), rads(c[1])
    >>> h2 = encode(*r, radians=True)
    >>> h == h2
    True
    """
    if radians:
        latitude = degrees(latitude)
        longitude = degrees(longitude)
    if longitude < 0:
        geoprint = ['w']
        loni = interval(-180.0, 0.0)
    else:
        geoprint = ['e']
        loni = interval(0.0, 180.0)

    lati = interval(-90.0, 90.0)

    while len(geoprint) < precision:
        ch = 0
        mid = (loni.min + loni.max) / 2
        if longitude > mid:
            ch |= 2
            loni = interval(mid, loni.max)
        else:
            loni = interval(loni.min, mid)

        mid = (lati.min + lati.max) / 2
        if latitude > mid:
            ch |= 1
            lati = interval(mid, lati.max)
        else:
            lati = interval(lati.min, mid)

        geoprint.append(alphabet[ch])
    return ''.join(geoprint)


def decode(geoprint, radians=False):
    """Decode a geoprint, returning the latitude and longitude.  These
    coordinates should approximate the input coordinates within a
    degree of error returned by 'error()'

    >>> c = (7.0625, -95.677068)
    >>> h = encode(*c)
    >>> c2 = decode(h)
    >>> e = error(h)
    >>> abs(c[0] - c2[0]) <= e
    True
    >>> abs(c[1] - c2[1]) <= e
    True

    If radians is True, results are in radians instead of degrees.

    >>> c2 = decode(h, radians=True)
    >>> e = error(h, radians=True)
    >>> abs(rads(c[0]) - c2[0]) <= e
    True
    >>> abs(rads(c[1]) - c2[1]) <= e
    True
    """
    lati = interval(-90.0, 90.0)
    first = geoprint[0]
    if first == 'w':
        loni = interval(-180.0, 0.0)
    elif first == 'e':
        loni = interval(0.0, 180.0)

    geoprint = geoprint[1:]

    for c in geoprint:
        cd = decodemap[c]
        if cd & 2:
            loni = interval((loni.min + loni.max) / 2, loni.max)
        else:
            loni = interval(loni.min, (loni.min + loni.max) / 2)
        if cd & 1:
            lati = interval((lati.min + lati.max) / 2, lati.max)
        else:
            lati = interval(lati.min, (lati.min + lati.max) / 2)
    lat = (lati.min + lati.max) / 2
    lon = (loni.min + loni.max) / 2
    if radians:
        return rads(lat), rads(lon)
    return lat, lon


def error(geoprint, radians=False):
    """Returns the error of a given geoprint.

    If radians is true, return the error in radians, otherwise
    degrees.
    """
    e = 90.0 * pow(2, neg(len(geoprint) - 1))
    if radians:
        return rads(e)
    return e


def size(geoprint):
    """Return the *approximate* size in meters of one side of a
    geoprint box.
    """
    return 20000000.0 * pow(2, neg(len(geoprint) - 1))


def distance(start, end, radians=False):
    """
    Calculate the *approximate* distance between two geoprints.  Based
    on Haversine formula
    (http://en.wikipedia.org/wiki/Haversine_formula).

    If radians is True, return distance is radians on a approximate
    great circle, otherwise return meters.
    """
    start_lat, start_lon = decode(start, radians=True)
    end_lat, end_lon = decode(end, radians=True)
    d_lat = end_lat - start_lat
    d_long = end_lon - start_lon
    a = (sin(d_lat / 2) ** 2 + cos(start_lat) *
         cos(end_lat) * sin(d_long / 2) ** 2)
    c = 2 * atan2(sqrt(a), sqrt(1 - a))
    if radians:
        return c
    return EARTH_RADIUS * c


def bearing(start, end, radians=False):
    """
    Calculate *approximate* bearing between two geoprints.

    If radians is True, returns bearing in radians, otherwise degrees.
    """
    start_lat, start_lon = decode(start, radians=True)
    end_lat, end_lon = decode(end, radians=True)
    d_lon = end_lon - start_lon
    c = atan2(sin(d_lon) * cos(end_lat),
              cos(start_lat) * sin(end_lat) -
              sin(start_lat) * cos(end_lat) * cos(d_lon))
    if radians:
        return c
    return degrees(c)


def format(geoprint):
    """Return two formatted strings of latitude and longitude,
    truncated to the number of decimal places known based on the
    computed error in the geoprint.
    """
    lat, lon = decode(geoprint)
    err = max(1, int(round(-log10(error(geoprint))))) - 1
    lats = "%.*f" % (err, lat)
    lons = "%.*f" % (err, lon)
    if '.' in lats:
        lats = lats.rstrip('0')
    if '.' in lons:
        lons = lons.rstrip('0')
    return lats, lons


_radials = {'N' : 0,              # N
            'NE': pi / 4,         # NE
            'E' : pi / 2,         # E
            'SE': (3 * pi / 4),   # SE
            'S' : pi,             # S
            'SW': (5 * pi / 4),   # SW
            'W' : (3 * pi / 2),   # W
            'NW': (7 * pi / 4),   # NW
            }


def neighbors(geoprint, bearing=True):
    """
    Return the eight neighboring geoprints and optionally their
    compass bearing flag around the given geoprint.

    If bearing is True, bearing flag will be either None (not
    adjacent) or one of ['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW']
    """
    results = set()
    precision = len(geoprint)
    lat1, lon1 = decode(geoprint, radians=True)
    d = size(geoprint) / EARTH_RADIUS
    for b, r in _radials.iteritems():
        lat = asin(sin(lat1) * cos(d) +
                   cos(lat1) * sin(d) * cos(r))
        dlon = atan2(sin(r) * sin(d) * cos(lat1),
                     cos(d) - sin(lat1) * sin(lat))
        lon = mod(lon1 + dlon + pi, 2 * pi) - pi
        h = encode(lat, lon, precision=precision, radians=True)
        if bearing:
            h = (b, h)
        results.add(h)
    return results


def adjacent(first, second):
    """
    Return a flag indicating if two geoprints are adjacent to each
    other.
    """
    if min(len(first), len(second)) < 3:
        raise TypeError(
            "Adjacency requires at least 3 characters of precision.")
    maxe = (error(first, True) / 2) + (error(second, True) / 2)
    maxe2 = pow(maxe, 2)
    h = sqrt(maxe2 + maxe2)
    d = distance(first, second, True)
    return h <= d



# -*- coding: utf-8 -*-

stopwords = """
 | A German stop word list. Comments begin with vertical bar. Each stop
 | word is at the start of a line.

 | The number of forms in this list is reduced significantly by passing it
 | through the German stemmer.


aber           |  but

alle           |  all
allem
allen
aller
alles

als            |  than, as
also           |  so
am             |  an + dem
an             |  at

ander          |  other
andere
anderem
anderen
anderer
anderes
anderm
andern
anderr
anders

auch           |  also
auf            |  on
aus            |  out of
bei            |  by
bin            |  am
bis            |  until
bist           |  art
da             |  there
damit          |  with it
dann           |  then

der            |  the
den
des
dem
die
das

daß            |  that

derselbe       |  the same
derselben
denselben
desselben
demselben
dieselbe
dieselben
dasselbe

dazu           |  to that

dein           |  thy
deine
deinem
deinen
deiner
deines

denn           |  because

derer          |  of those
dessen         |  of him

dich           |  thee
dir            |  to thee
du             |  thou

dies           |  this
diese
diesem
diesen
dieser
dieses


doch           |  (several meanings)
dort           |  (over) there


durch          |  through

ein            |  a
eine
einem
einen
einer
eines

einig          |  some
einige
einigem
einigen
einiger
einiges

einmal         |  once

er             |  he
ihn            |  him
ihm            |  to him

es             |  it
etwas          |  something

euer           |  your
eure
eurem
euren
eurer
eures

für            |  for
gegen          |  towards
gewesen        |  p.p. of sein
hab            |  have
habe           |  have
haben          |  have
hat            |  has
hatte          |  had
hatten         |  had
hier           |  here
hin            |  there
hinter         |  behind

ich            |  I
mich           |  me
mir            |  to me


ihr            |  you, to her
ihre
ihrem
ihren
ihrer
ihres
euch           |  to you

im             |  in + dem
in             |  in
indem          |  while
ins            |  in + das
ist            |  is

jede           |  each, every
jedem
jeden
jeder
jedes

jene           |  that
jenem
jenen
jener
jenes

jetzt          |  now
kann           |  can

kein           |  no
keine
keinem
keinen
keiner
keines

können         |  can
könnte         |  could
machen         |  do
man            |  one

manche         |  some, many a
manchem
manchen
mancher
manches

mein           |  my
meine
meinem
meinen
meiner
meines

mit            |  with
muss           |  must
musste         |  had to
nach           |  to(wards)
nicht          |  not
nichts         |  nothing
noch           |  still, yet
nun            |  now
nur            |  only
ob             |  whether
oder           |  or
ohne           |  without
sehr           |  very

sein           |  his
seine
seinem
seinen
seiner
seines

selbst         |  self
sich           |  herself

sie            |  they, she
ihnen          |  to them

sind           |  are
so             |  so

solche         |  such
solchem
solchen
solcher
solches

soll           |  shall
sollte         |  should
sondern        |  but
sonst          |  else
über           |  over
um             |  about, around
und            |  and

uns            |  us
unse
unsem
unsen
unser
unses

unter          |  under
viel           |  much
vom            |  von + dem
von            |  from
vor            |  before
während        |  while
war            |  was
waren          |  were
warst          |  wast
was            |  what
weg            |  away, off
weil           |  because
weiter         |  further

welche         |  which
welchem
welchen
welcher
welches

wenn           |  when
werde          |  will
werden         |  will
wie            |  how
wieder         |  again
will           |  want
wir            |  we
wird           |  will
wirst          |  willst
wo             |  where
wollen         |  want
wollte         |  wanted
würde          |  would
würden         |  would
zu             |  to
zum            |  zu + dem
zur            |  zu + der
zwar           |  indeed
zwischen       |  between

"""
